//===========================================================
//=     University of Illinois at Urbana/Champaign          =
//=     Department of Computer Science                      =
//=     Dr. Dan Roth - Cognitive Computation Group          =
//=                                                         =
//=  Project: Snow                                          =
//=                                                         =
//=   Module: Network.cpp                                   =
//=  Version: 2.1                                           =
//=   Author: Jeff Rosen, Andrew Carlson                    =
//=     Date: xx/xx/99                                      = 
//=                                                         =
//= Comments:                                               =
//===========================================================

#include "GlobalParams.h"
#include "Network.h"
#include "Winnow.h"
#include "Perceptron.h"
#include "NaiveBayes.h"
#include "Cloud.h"
#include <string>
#include <typeinfo>
#include <algorithm>

bool Network::CreateStructure( const string& spec )
{
  bool result = true;
  
  char* workSpec = new char[spec.length() + 1];
  if (workSpec != NULL)
    {
      // Make a working copy
      strcpy(workSpec, spec.c_str());
      if (globalParams.verbosity != VERBOSE_QUIET) 
        { 
	  cout << "Network Spec -> " << workSpec << endl;
        } 
      char* token = strtok(workSpec, ":(, \t\n\r");
      char* delim;
      int index = 0; 

      while (token != NULL)
        {
	  // token must now be an algorithm type
	  if (tolower(*token) == 'w')
            {
	      double alpha = 1.35;
	      double beta = 0.8;
	      double threshold = 4.0;
	      double defaultWeight = 0.2;

	      token = token + strlen(token) + 1;
	      // if only a target set is specified, keep defaults
	      if ((*token) == ':')
                {
		  delim = token; 
                } else {
		  alpha = strtod(token, &delim);
		  token = delim + strspn(delim, ":, \t\n\r");
		  beta = strtod(token, &delim);
		  token = delim + strspn(delim, ":, \t\n\r");
		  threshold = strtod(token, &delim);
		  token = delim + strspn(delim, ":, \t\n\r");
		  defaultWeight = strtod(token, &delim);
                }
                // make sure a target set was specified
                if (*delim == ':') 
		  {
                    token = delim + strspn(delim, ":), \t\n\r"); 
                    token = strtok(token, ")"); 
                    delim+=strlen(token)+2;                
		  } else {
                    delim+=strspn(delim, ":) \t\n\r"); 
                    cerr << "Error: No target set given for Winnow" << endl;
                    return false;
		  }
                if (*delim == ',')
		  {
                    Winnow* pWin =
		      new Winnow(alpha, beta, threshold, defaultWeight);
                    pWin->targetIds.Parse(token); 
                    pWin->index = index; 
                    TargetIdSet::iterator it = pWin->targetIds.begin();
                    TargetIdSet::iterator end = pWin->targetIds.end();
                    for ( ; it != end ; it++) 
		      {
                        if (globalParams.targetIds.find(*it) ==
                            globalParams.targetIds.end())
			  {
                            globalParams.targetIds.insert(*it);
			  }
		      } 
                    algorithms.push_back(pWin);
		  } else {
                    delim = token;
                    token = strtok(NULL, ")");
                    cerr << "Failed to parse winnow spec '" <<
		      delim << '(' << token << ')' <<
		      endl << endl;
                    result = false;
		  }        
            } else if (tolower(*token) == 'p')
	      {
                token = token + strlen(token) + 1;
                
                double learningRate = 0.1;
                double threshold = 4.0;
                double defaultWeight = 0.2; 
                
                if ((*token) == ':')
		  {
                    delim = token;
		  } else {
                    learningRate = strtod(token, &delim);
                    token = delim + strspn(delim, ", \t\n\r");
                    threshold = strtod(token, &delim);
                    token = delim + strspn(delim, ", \t\n\r");
                    defaultWeight = strtod(token, &delim);
		  }
                if (*delim == ':') 
		  {
                    token = delim + strspn(delim, ":), \t\n\r"); 
                    token = strtok(token, ")"); 
                    delim+=strlen(token)+2;                
		  } else {
                    delim+=strspn(delim, ":) \t\n\r"); 
                    cerr << "Error: No target set given for Perceptron" << endl;
                    return false;
		  }
                if (*delim == ',') 
		  {
                    Perceptron* pPer =
		      new Perceptron(learningRate, threshold, defaultWeight);
                    pPer->targetIds.Parse(token); 
                    pPer->index = index;
                    TargetIdSet::iterator it = pPer->targetIds.begin();
                    TargetIdSet::iterator end = pPer->targetIds.end();
                    for ( ; it != end ; it++) 
		      {
                        if (globalParams.targetIds.find(*it) ==
                            globalParams.targetIds.end())
			  {
                            globalParams.targetIds.insert(*it);
			  }
		      } 
                    algorithms.push_back(pPer);
		  } else {
                    delim = token;
                    token = strtok(NULL, ")");
                    cerr << "Failed to parse perceptron spec '" <<
		      delim << '(' << token << ')' <<
		      endl << endl;
                    result = false;
		  }        
	      } else if (tolower(*token) == 'b')
		{
		  delim = token + strlen(token) + 1;
		  token = delim + strspn(delim, ":), \t\n\r"); 
		  token = strtok(token, ")"); 
		  delim+=strlen(token)+2;                
		  if (*delim == ',')
		    {
		      NaiveBayes* pBay =
                        new NaiveBayes();
		      pBay->targetIds.Parse(token); 
		      pBay->index = index;
		      TargetIdSet::iterator it = pBay->targetIds.begin();
		      TargetIdSet::iterator end = pBay->targetIds.end();
		      for ( ; it != end ; it++) 
			{
			  if (globalParams.targetIds.find(*it) ==
			      globalParams.targetIds.end())
			    {
			      globalParams.targetIds.insert(*it);
			    }
			} 
		      algorithms.push_back(pBay);
		    } else {
		      delim = token;
		      token = strtok(NULL, ")");
		      cerr << "Failed to parse naive Bayes spec '" <<
                        delim << '(' << token << ')' <<
                        endl << endl;
		      result = false;
		    } 
		}
	  index++;

	  token = strtok(delim, " (),\t\n\r");
        } 
      
      CreateClouds(globalParams.targetIds, algorithms);
      
    } else {
      cerr << "Error:" << endl;
      cerr << "Failed to allocate work buffer for parsing " <<
	"network specification" << endl;
      result = false;
    }
    return result;
}

void Network::CreateClouds( TargetIdSet& tars, AlgorithmVector& algorithms )       
{
  AlgorithmVector activeAlgorithms;
  
  TargetIdSet::const_iterator it = tars.begin();
  TargetIdSet::const_iterator end = tars.end();
  
  // Create one new Cloud for each TargetId
  for ( ; it != end; it++)
    {
      activeAlgorithms.clear();
      AlgorithmVector::iterator algIt = algorithms.begin();
      AlgorithmVector::iterator algEnd = algorithms.end();
      for ( ; algIt != algEnd; algIt++) 
        {
	  if ( (*algIt)->targetIds.find(*it) !=
	       ( (*algIt)->targetIds.end() ) )
            {
	      activeAlgorithms.push_back(*algIt);
            }    
        }
      clouds.push_back(Cloud(*it, activeAlgorithms));
    }
}
    
bool Network::PresentExample( Example& ex )
{
  Example::FeatureMap::const_iterator feat = ex.GetFeatureBegin();
  Example::FeatureMap::const_iterator last = ex.GetFeatureEnd();
  
  // Update the globalFeature counts if necessary
  if (globalParams.eligibilityThreshold > 1)
    {
      for ( ; feat != last; feat++)
        {
	  FeatureCountMap::iterator fc;
	  if ((fc = globalFeatureCount.find(feat->first))
	      != globalFeatureCount.end())
            {
	      // Increment the count
	      fc->second++;
            } else {
	      // Initialize the count to 1
	      globalFeatureCount.insert(make_pair(feat->first, 1));
            }
        }
    }
  
  CloudVector::iterator it = clouds.begin();
  CloudVector::iterator end = clouds.end();
  bool result = false;
  
  for ( ; it != end; it++)
    {
      if (it->PresentExample(ex))
        {
	  result = true;
        }
    }
  
  return result;
}

void Network::TrainingComplete()
{
  // This is only necessary for Naive Bayes
  NaiveBayes nb;
  AlgorithmVector::iterator algIt = algorithms.begin();
  AlgorithmVector::iterator algEnd = algorithms.end();
  for ( ; algIt != algEnd; algIt++)	
    if (typeid(nb) == typeid(**algIt))
      dynamic_cast<NaiveBayes*>(*algIt)->ClearTotalPrior();
  
  // Walk through clouds
  CloudVector::iterator it = clouds.begin();
  CloudVector::iterator end = clouds.end();
  for ( ; it != end; it++)
    {
      it->TrainingComplete();
    }  
}

void Network::Discard()
{
  if (globalParams.discardMethod != DISCARD_NONE)
    {   
      if (globalParams.discardMethod == DISCARD_ABS)
	{ 
	  CloudVector::iterator it = clouds.begin();
	  CloudVector::iterator end = clouds.end();
	  for ( ; it != end; it++)
	    {
	      it->Discard();
	    }
	}
      else {
	// NOTE: This method really only makes sense for Winnow.
	//       We'll allow it for others but it probably won't
	//       yield good results for others.
            
	CloudVector::iterator it = clouds.begin();
	CloudVector::iterator end = clouds.end();
	set<FeatureID> allFeatures;
	for ( ; it != end; it++)
	  {
	    it->CollectFeatures(allFeatures);
	  }
			
	set<FeatureID>::const_iterator feat;
	for (feat = allFeatures.begin();
	     feat != allFeatures.end();
	     feat++)
	  {
	    it = clouds.begin();
	    double min = 1.0e300;
	    int dups = 0;
	    for (it = clouds.begin(); it != end; it++)
	      {
		it->FindMinimumWeight(*feat, min, dups);
	      }
	    
	    if (!dups)
	      {
		for (it = clouds.begin(); it != end; it++)
		  {
		    it->Discard(*feat, min);
		  }
	      }
	  }
      }
    }        
}


void Network::RankTargets( Example& ex, TargetRanking& rank )
{
  AlgorithmVector::iterator algIt = algorithms.begin();
  AlgorithmVector::iterator algEnd = algorithms.end();
  
  for ( ; algIt != algEnd; algIt++)
    (*algIt)->PrepareToRank();
  
  CloudVector::iterator it = clouds.begin();
  CloudVector::iterator end = clouds.end();
  
  for ( ; it != end; it++)
    it->PreparetoRank(ex); 
    
  double cloudActivation;
  
  for ( it = clouds.begin() ; it != end; it++)
    {
      // make sure cloud is in the targetIDSet
      if (globalParams.targetIds.find(it->Id()) !=
	  globalParams.targetIds.end())
        {
	  // get the activation for the current cloud and 
	  // insert it into the ranking
	  cloudActivation = it->RankTargets(ex);
	  TargetRank tr(it->Id(), cloudActivation);
	  rank.push_back(tr);
        }
    }
}

void Network::Show()
{
  AlgorithmVector::iterator algIt = algorithms.begin();
  AlgorithmVector::iterator algEnd = algorithms.end();
  
  for ( ; algIt != algEnd; algIt++)
    {
      (*algIt)->Show(cout);
    }

  CloudVector::iterator it = clouds.begin();
  CloudVector::iterator end = clouds.end();
  for ( ; it != end; it++)
    {
      it->Show();
    }
}

void Network::Read( ifstream& in )
{
  string Flag;
  Target* tempTarget;
  
  CloudVector::iterator cloudIt = clouds.begin();
  CloudVector::iterator cloudEnd = clouds.end();
 
  while (!in.fail())
    {  
      operator>>(in, Flag);
                
      if ((Flag == "target") && (!in.fail()))
        {
	  tempTarget = new Target;
	  tempTarget->Read(in, algorithms);
	  globalParams.targetIds.insert(tempTarget->Id());
            
	  // step through Clouds, looking for a matching targetID
	  for (cloudIt = clouds.begin(), cloudEnd = clouds.end(); 
	       (cloudIt != cloudEnd) && (cloudIt->Id() != tempTarget->Id()); 
	       cloudIt++);
	  if (cloudIt == cloudEnd) {
	    // no Cloud exists with the correct targetID, so create one
	    // and add the Target to it 
	    clouds.push_back(Cloud(tempTarget->Id()));
	    clouds.back().AddTarget(*tempTarget);
	  } else {  
	    // we've found a Cloud with the correct targetID, so just
	    // add the new Target to it
	    cloudIt->AddTarget(*tempTarget);
	  }
        }
    } 
}

void Network::Write( ofstream& out )
{
    // Walk through clouds and write them out
    CloudVector::iterator it = clouds.begin();
    CloudVector::iterator end = clouds.end();
    for ( ; it != end; it++)
    {
        it->Write(out);
        out << endl;
    }
}


